/*
    Classe CJarvis
*/

#include <math.h>
#include "CJarvis.h"

CJarvis::CJarvis() {
   start = 0;
}

// traz o primeiro ponto jarvis
// estratgia: procura o ponto mais a esquerda
CCoord* CJarvis::first() {
   CCoord *n = points.first();
   start = n;
   while(n) {
      n->jarvis = false;
      if(n->x < start->x) {
          points.mark();
          start = n;
      }
      n = points.next();
   }
   if(start)
      start->jarvis = true;
   return start;
}

// traz o ltimo ponto jarvis
// estratgia: percorre todos os pontos usando o algoritmo jarvis
//             e retorna o ltimo
CCoord* CJarvis::last() {
   CCoord *n = first(), *c = 0;
   while(n) {
      c = n;
      n = next();
   }
   return c;
}

// traz o prximo ponto jarvis
// estratgia: compara o ponto atual com os ainda no testados
//             e retorna o ponto de menor angulo
CCoord* CJarvis::next() {
   CCoord *c = 0, *n, *a;
   float ang, aang;
   
   if(start) {
       points.restore();
       a = points.get();
       if(a) {
           aang = 360;
           n = points.first();
           while(n) {
               if(!n->jarvis) {
                   ang = calcAngle(a, n);
                   if(ang < aang) {
                       points.mark();
                       c = n;
                       aang = ang;
                   }
               }
               n = points.next();
           }
       }
       if(c == start) 
            start = 0;            // se fechou o poligono, tem que encerrar
       else if(c) 
            c->jarvis = true;     // marca esse ponto como j processado
            
       if(start)
          start->jarvis = false;  // fora que o ponto de inicio seja testado na proxima vez
   }      
   
   return c;
}
     
// calcula o angulo entre dois pontos
// referencia: http://www.chrisharrison.net/projects/convexHull/index.html
float CJarvis::calcAngle(CCoord *p1, CCoord *p2) {
    float deltaX = p2->x - p1->x;
    float deltaY = p2->y - p1->y;
    float angle;
    
    if (deltaX==0 && deltaY==0)
        return 0;
    
    angle = atan(deltaY/deltaX)*(180.0/3.1415926);

    //verifica os quadrantes, valores possiveis: 0-360
    if (deltaX>=0 && deltaY>=0)
        angle += 90;
    else if (deltaX>=0 && deltaY<0)
        angle += 90;
    else if (deltaX<0 && deltaY>0)
        angle += 270;
    else if (deltaX<0 && deltaY<=0)
        angle += 270;
    
    return angle;
//     return (atan2(p2->y - p1->y, p2->x - p1->x)*180.0)/3.1415926;
}

/*
  2D math 
  referencias: http://www.geocities.com/SiliconValley/2151/math2d.html

- The 2D Dot Product

The 2D dot product between two 2D vectors <Ax,Ay> and <Bx,By> is calculated like so:

A * B = Ax*Bx + Ay*By
If A and B are unit vectors then the dot product is the cosine of the angle between them, so the angle itself (theta) can be calculated by taking the inverse cosine of the dot product:

theta = invcos(A * B)
Fortunately you'll often only need the cosine of the angle between 2 vectors and not the angle itself, so the expensive step of calculating the inverse cosine can be skipped.

- 2D Rotation

A point <x,y> can be rotated around the origin <0,0> by running it through the following equations to get the new point <x',y'> :

x' = cos(theta)*x - sin(theta)*y 
y' = sin(theta)*x + cos(theta)*y
where theta is the angle by which to rotate the point.

- Clockwise/Anti-clockwise

The dot product and rotation equations give us a handy way of determining whether the points in a given polygon are stored in clockwise or anti-clockwise fashion.

Thus we can use the following formula to determine if the 3 points are in clockwise order:
E1x = P1x-P2x 
E1y = P1y-P2y 
E2x = P3x-P2x 
E2y = P3y-P2y
if ((E1x * E2y - E1y * E2x) >= 0) clockwise = TRUE; 
else clockwise = FALSE;

*/

